/**
 * Copyright (c) 2021 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Sukolsak Sakshuwong <sukolsak@stanford.edu>
 */
import { __awaiter, __extends, __generator } from "tslib";
import { sort, arraySwap } from '../../mol-data/util';
import { asciiWrite } from '../../mol-io/common/ascii';
import { Vec3, Mat3, Mat4 } from '../../mol-math/linear-algebra';
import { StringBuilder } from '../../mol-util';
import { Color } from '../../mol-util/color/color';
import { zip } from '../../mol-util/zip/zip';
import { MeshExporter } from './mesh-exporter';
// avoiding namespace lookup improved performance in Chrome (Aug 2020)
var v3fromArray = Vec3.fromArray;
var v3transformMat4 = Vec3.transformMat4;
var v3transformMat3 = Vec3.transformMat3;
var mat3directionTransform = Mat3.directionTransform;
var ObjExporter = /** @class */ (function (_super) {
    __extends(ObjExporter, _super);
    function ObjExporter(filename) {
        var _this = _super.call(this) || this;
        _this.filename = filename;
        _this.fileExtension = 'zip';
        _this.obj = StringBuilder.create();
        _this.mtl = StringBuilder.create();
        _this.vertexOffset = 0;
        _this.materialSet = new Set();
        StringBuilder.writeSafe(_this.obj, "mtllib " + filename + ".mtl\n");
        return _this;
    }
    ObjExporter.prototype.updateMaterial = function (color, alpha) {
        if (this.currentColor === color && this.currentAlpha === alpha)
            return;
        this.currentColor = color;
        this.currentAlpha = alpha;
        var material = Color.toHexString(color) + alpha;
        StringBuilder.writeSafe(this.obj, "usemtl " + material);
        StringBuilder.newline(this.obj);
        if (!this.materialSet.has(material)) {
            this.materialSet.add(material);
            var _a = Color.toRgbNormalized(color), r = _a[0], g = _a[1], b = _a[2];
            var mtl = this.mtl;
            StringBuilder.writeSafe(mtl, "newmtl " + material + "\n");
            StringBuilder.writeSafe(mtl, 'illum 2\n'); // illumination model
            StringBuilder.writeSafe(mtl, 'Ns 163\n'); // specular exponent
            StringBuilder.writeSafe(mtl, 'Ni 0.001\n'); // optical density a.k.a. index of refraction
            StringBuilder.writeSafe(mtl, 'Ka 0 0 0\n'); // ambient reflectivity
            StringBuilder.writeSafe(mtl, 'Kd '); // diffuse reflectivity
            StringBuilder.writeFloat(mtl, r, 1000);
            StringBuilder.whitespace1(mtl);
            StringBuilder.writeFloat(mtl, g, 1000);
            StringBuilder.whitespace1(mtl);
            StringBuilder.writeFloat(mtl, b, 1000);
            StringBuilder.newline(mtl);
            StringBuilder.writeSafe(mtl, 'Ks 0.25 0.25 0.25\n'); // specular reflectivity
            StringBuilder.writeSafe(mtl, 'd '); // dissolve
            StringBuilder.writeFloat(mtl, alpha, 1000);
            StringBuilder.newline(mtl);
        }
    };
    ObjExporter.quantizeColors = function (colorArray, vertexCount) {
        if (vertexCount <= 1024)
            return;
        var rgb = Vec3();
        var min = Vec3();
        var max = Vec3();
        var sum = Vec3();
        var colorMap = new Map();
        var colorComparers = [
            function (colors, i, j) { return (Color.toVec3(rgb, colors[i])[0] - Color.toVec3(rgb, colors[j])[0]); },
            function (colors, i, j) { return (Color.toVec3(rgb, colors[i])[1] - Color.toVec3(rgb, colors[j])[1]); },
            function (colors, i, j) { return (Color.toVec3(rgb, colors[i])[2] - Color.toVec3(rgb, colors[j])[2]); },
        ];
        var medianCut = function (colors, l, r, depth) {
            if (l > r)
                return;
            if (l === r || depth >= 10) {
                // Find the average color.
                Vec3.set(sum, 0, 0, 0);
                for (var i = l; i <= r; ++i) {
                    Color.toVec3(rgb, colors[i]);
                    Vec3.add(sum, sum, rgb);
                }
                Vec3.round(rgb, Vec3.scale(rgb, sum, 1 / (r - l + 1)));
                var averageColor = Color.fromArray(rgb, 0);
                for (var i = l; i <= r; ++i)
                    colorMap.set(colors[i], averageColor);
                return;
            }
            // Find the color channel with the greatest range.
            Vec3.set(min, 255, 255, 255);
            Vec3.set(max, 0, 0, 0);
            for (var i = l; i <= r; ++i) {
                Color.toVec3(rgb, colors[i]);
                for (var j = 0; j < 3; ++j) {
                    Vec3.min(min, min, rgb);
                    Vec3.max(max, max, rgb);
                }
            }
            var k = 0;
            if (max[1] - min[1] > max[k] - min[k])
                k = 1;
            if (max[2] - min[2] > max[k] - min[k])
                k = 2;
            sort(colors, l, r + 1, colorComparers[k], arraySwap);
            var m = (l + r) >> 1;
            medianCut(colors, l, m, depth + 1);
            medianCut(colors, m + 1, r, depth + 1);
        };
        // Create an array of unique colors and use the median cut algorithm.
        var colorSet = new Set();
        for (var i = 0; i < vertexCount; ++i) {
            colorSet.add(Color.fromArray(colorArray, i * 3));
        }
        var colors = Array.from(colorSet);
        medianCut(colors, 0, colors.length - 1, 0);
        // Map actual colors to quantized colors.
        for (var i = 0; i < vertexCount; ++i) {
            var color = colorMap.get(Color.fromArray(colorArray, i * 3));
            Color.toArray(color, colorArray, i * 3);
        }
    };
    ObjExporter.prototype.addMeshWithColors = function (input) {
        return __awaiter(this, void 0, void 0, function () {
            var mesh, values, isGeoTexture, webgl, ctx, obj, t, n, tmpV, stride, groupCount, colorType, tColor, uAlpha, dTransparency, tTransparency, aTransform, instanceCount, interpolatedColors, instanceIndex, _a, vertices, normals, indices, groups, vertexCount, drawCount, i, i, i, color, group, group, alpha, group, transparency, v1, v2, v3;
            return __generator(this, function (_b) {
                switch (_b.label) {
                    case 0:
                        mesh = input.mesh, values = input.values, isGeoTexture = input.isGeoTexture, webgl = input.webgl, ctx = input.ctx;
                        obj = this.obj;
                        t = Mat4();
                        n = Mat3();
                        tmpV = Vec3();
                        stride = isGeoTexture ? 4 : 3;
                        groupCount = values.uGroupCount.ref.value;
                        colorType = values.dColorType.ref.value;
                        tColor = values.tColor.ref.value.array;
                        uAlpha = values.uAlpha.ref.value;
                        dTransparency = values.dTransparency.ref.value;
                        tTransparency = values.tTransparency.ref.value;
                        aTransform = values.aTransform.ref.value;
                        instanceCount = values.uInstanceCount.ref.value;
                        if (colorType === 'volume' || colorType === 'volumeInstance') {
                            interpolatedColors = ObjExporter.getInterpolatedColors(mesh.vertices, mesh.vertexCount, values, stride, colorType, webgl);
                            ObjExporter.quantizeColors(interpolatedColors, mesh.vertexCount);
                        }
                        return [4 /*yield*/, ctx.update({ isIndeterminate: false, current: 0, max: instanceCount })];
                    case 1:
                        _b.sent();
                        instanceIndex = 0;
                        _b.label = 2;
                    case 2:
                        if (!(instanceIndex < instanceCount)) return [3 /*break*/, 6];
                        if (!ctx.shouldUpdate) return [3 /*break*/, 4];
                        return [4 /*yield*/, ctx.update({ current: instanceIndex + 1 })];
                    case 3:
                        _b.sent();
                        _b.label = 4;
                    case 4:
                        _a = ObjExporter.getInstance(input, instanceIndex), vertices = _a.vertices, normals = _a.normals, indices = _a.indices, groups = _a.groups, vertexCount = _a.vertexCount, drawCount = _a.drawCount;
                        Mat4.fromArray(t, aTransform, instanceIndex * 16);
                        mat3directionTransform(n, t);
                        // position
                        for (i = 0; i < vertexCount; ++i) {
                            v3transformMat4(tmpV, v3fromArray(tmpV, vertices, i * stride), t);
                            StringBuilder.writeSafe(obj, 'v ');
                            StringBuilder.writeFloat(obj, tmpV[0], 1000);
                            StringBuilder.whitespace1(obj);
                            StringBuilder.writeFloat(obj, tmpV[1], 1000);
                            StringBuilder.whitespace1(obj);
                            StringBuilder.writeFloat(obj, tmpV[2], 1000);
                            StringBuilder.newline(obj);
                        }
                        // normal
                        for (i = 0; i < vertexCount; ++i) {
                            v3transformMat3(tmpV, v3fromArray(tmpV, normals, i * stride), n);
                            StringBuilder.writeSafe(obj, 'vn ');
                            StringBuilder.writeFloat(obj, tmpV[0], 100);
                            StringBuilder.whitespace1(obj);
                            StringBuilder.writeFloat(obj, tmpV[1], 100);
                            StringBuilder.whitespace1(obj);
                            StringBuilder.writeFloat(obj, tmpV[2], 100);
                            StringBuilder.newline(obj);
                        }
                        // face
                        for (i = 0; i < drawCount; i += 3) {
                            color = void 0;
                            switch (colorType) {
                                case 'uniform':
                                    color = Color.fromNormalizedArray(values.uColor.ref.value, 0);
                                    break;
                                case 'instance':
                                    color = Color.fromArray(tColor, instanceIndex * 3);
                                    break;
                                case 'group': {
                                    group = isGeoTexture ? ObjExporter.getGroup(groups, i) : groups[indices[i]];
                                    color = Color.fromArray(tColor, group * 3);
                                    break;
                                }
                                case 'groupInstance': {
                                    group = isGeoTexture ? ObjExporter.getGroup(groups, i) : groups[indices[i]];
                                    color = Color.fromArray(tColor, (instanceIndex * groupCount + group) * 3);
                                    break;
                                }
                                case 'vertex':
                                    color = Color.fromArray(tColor, indices[i] * 3);
                                    break;
                                case 'vertexInstance':
                                    color = Color.fromArray(tColor, (instanceIndex * vertexCount + indices[i]) * 3);
                                    break;
                                case 'volume':
                                    color = Color.fromArray(interpolatedColors, (isGeoTexture ? i : indices[i]) * 3);
                                    break;
                                case 'volumeInstance':
                                    color = Color.fromArray(interpolatedColors, (instanceIndex * vertexCount + (isGeoTexture ? i : indices[i])) * 3);
                                    break;
                                default: throw new Error('Unsupported color type.');
                            }
                            alpha = uAlpha;
                            if (dTransparency) {
                                group = isGeoTexture ? ObjExporter.getGroup(groups, i) : groups[indices[i]];
                                transparency = tTransparency.array[instanceIndex * groupCount + group] / 255;
                                alpha *= 1 - transparency;
                            }
                            this.updateMaterial(color, alpha);
                            v1 = this.vertexOffset + (isGeoTexture ? i : indices[i]) + 1;
                            v2 = this.vertexOffset + (isGeoTexture ? i + 1 : indices[i + 1]) + 1;
                            v3 = this.vertexOffset + (isGeoTexture ? i + 2 : indices[i + 2]) + 1;
                            StringBuilder.writeSafe(obj, 'f ');
                            StringBuilder.writeInteger(obj, v1);
                            StringBuilder.writeSafe(obj, '//');
                            StringBuilder.writeIntegerAndSpace(obj, v1);
                            StringBuilder.writeInteger(obj, v2);
                            StringBuilder.writeSafe(obj, '//');
                            StringBuilder.writeIntegerAndSpace(obj, v2);
                            StringBuilder.writeInteger(obj, v3);
                            StringBuilder.writeSafe(obj, '//');
                            StringBuilder.writeInteger(obj, v3);
                            StringBuilder.newline(obj);
                        }
                        this.vertexOffset += vertexCount;
                        _b.label = 5;
                    case 5:
                        ++instanceIndex;
                        return [3 /*break*/, 2];
                    case 6: return [2 /*return*/];
                }
            });
        });
    };
    ObjExporter.prototype.getData = function () {
        return {
            obj: StringBuilder.getString(this.obj),
            mtl: StringBuilder.getString(this.mtl)
        };
    };
    ObjExporter.prototype.getBlob = function (ctx) {
        return __awaiter(this, void 0, void 0, function () {
            var _a, obj, mtl, objData, mtlData, zipDataObj, _b;
            var _c;
            return __generator(this, function (_d) {
                switch (_d.label) {
                    case 0:
                        _a = this.getData(), obj = _a.obj, mtl = _a.mtl;
                        objData = new Uint8Array(obj.length);
                        asciiWrite(objData, obj);
                        mtlData = new Uint8Array(mtl.length);
                        asciiWrite(mtlData, mtl);
                        zipDataObj = (_c = {},
                            _c[this.filename + '.obj'] = objData,
                            _c[this.filename + '.mtl'] = mtlData,
                            _c);
                        _b = Blob.bind;
                        return [4 /*yield*/, zip(ctx, zipDataObj)];
                    case 1: return [2 /*return*/, new (_b.apply(Blob, [void 0, [_d.sent()], { type: 'application/zip' }]))()];
                }
            });
        });
    };
    return ObjExporter;
}(MeshExporter));
export { ObjExporter };
//# sourceMappingURL=obj-exporter.js.map